"""Game record types for shogi."""

from rshogi._rshogi import (
    GameRecord,
    GameRecordMetadata,
    GameRecordResult,
    MoveRecord,
    TimeControl,
    GameResult,
)

__all__ = [
    "GameRecord",
    "GameRecordMetadata",
    "GameRecordResult",
    "MoveRecord",
    "TimeControl",
    "GameResult",
]
